#ifndef __DECLARAT_H__

#define __DECLARAT_H__

#include <stdlib.h>
#include <unistd.h>

#include "macros.h"
#include "petscksp.h"
#include "petscda.h"
#include "petsc.h"

#if defined(MICSC_HAVE_SILO)
#include "silo.h"
#endif

/*E
  St_InputBcond - Structure containing the definition of a boundary
  condition as it is specified in the input grid file.

  Attributes:
+ x, y, z - Coordinates of the lower corner of the region where
            the boundary condition is defined.
. nx, ny, xz - Number of cells in each direction of the region where the
               boundary condition is defined.
. ineighb - Neighbor index.
. type - Type of boundary condition (see BcondType).
. porosity - Porosity value for the boundary condition.
- next - Pointer to the next structure in a list.

  Level: advanced
E*/
typedef struct St_InputBcond {
	PetscInt x, y, z, nx, ny, nz, ineighb;
	BcondType type;
	PetscReal porosity;
	struct St_InputBcond *next;
} St_InputBcond;

/*E
  St_Input - Structure containing all the input parameters of the problem,
  specified by the user (or set to default values).

  Attributes:
+ log_level - Logging level
. log_ievery - Number of iterations performed between logs
. out_ievery - Number of iterations performed between
                 binary solution outputs
. out_idtevery - Number of time steps performed between ascii
                   solution outputs
. startInstantMeans - First iteration for instant means calculation
. startVarianceMeans - First iteration for variance means calculation
. initial_field - Flag to print the initial field
. final_field - Flag to print the final field
. printProps - Flag to print properties
. max_outer_its - Maximum number of iterations
. min_outer_its - Minimum number of outer iterations
. resnorm_max - Maximum normalized residue
. corrnorm_max - Maximum normalized correction vector
. k_restart - Value to determine the restarting ('none', 'variables',
              'instant_means' or 'variance_means'). See RestartType.
. init_it - First iteration number
. k_wrap - Grid periodicity
. n_ghost - Ghost stencil width
. gravity - Gravity acceleration
. n_order_trans - Order of the transient scheme
. deltat - Time step
. tinit - Initial time
. tfinal - Final time
. uref - Constant value por the U direction inlet
. mu - MU value
. tref - TREF value
. rho - RHO value
. cs0 - LES constant value
. dpdx - Use DPDX property
. ndim - Number of dimensions
. nvar - Number of variables
. ncoupvar - Number of coupled variables
. contEqMethod - Continuity equation solving method
. inletFunction - Inlet function
. useLES - Flag to determine if the LES model should be used.
. useKEMOD - Flag to determine if the KE model should be used.
. useT - Flag to determine if the temperature variable should be used.
. firstBcond - Pointer to the first boundary condition of the list.
. firstBcond - Pointer to the first boundary condition of the list.
- lastBcond - Pointer to the last boundary condition of the list.

  Level: advanced
E*/
typedef struct {
	PetscInt log_level, log_ievery, out_ievery, out_idtevery;
	PetscInt startInstantMeans, startVarianceMeans;
	PetscTruth initial_field, final_field;
	PetscTruth printProps;
	PetscInt max_outer_its, min_outer_its;
	PetscReal resnorm_max, corrnorm_max;
	RestartType k_restart;
	PetscInt init_it;
	PetscInt k_wrap;
	PetscInt n_ghost;
	PetscReal gravity;
	PetscInt n_order_trans;
	PetscReal deltat, tinit, tfinal;
	PetscReal uref, mu, tref, rho, cs0;
	PetscTruth dpdx;
	PetscInt ndim, nvar, ncoupvar;
	PetscInt contEqMethod;
	PetscErrorCode (*inletFunction)(PetscInt*, PetscReal*);
	PetscTruth useLES, useKEMOD, useT;
	St_InputBcond *firstBcond, *lastBcond;
} St_Input;

/*E
  St_Grid - Structure for the grid.

  Attributes:
+ Nnode - Number of nodes (in each direction) of the global grid.
. Xnode - Coordinates of the nodes of the global grid.
- ktype - Grid type (see ktypeGrid).

  Level: advanced
E*/
typedef struct {
	PetscInt Nnode[3];
	PetscReal *Xnode[3];
	ktypeGrid ktype;
} St_Grid;

/*E
  St_Trans - Structure with parameters for solving unsteady problems.

  Attributes:
+ norder - Transient order (see TransientOrder).
. time - Current time.
- idt - Current time step.

  Level: advanced
E*/
typedef struct {
	TransientOrder norder;
	PetscReal time;
	PetscInt idt;
} St_Trans;

/*E
  MeshRegion - Structure containing the attributes of a mesh region.

  Attributes:
+ nCells - Number of cells in each direction.
. lowerCorner - Coordinate of the lower corner of the mesh region.
. upperCorner - Coordinate of the upper corner of the mesh region.
- totalCells - Total number of cells.

  Level: advanced
E*/
typedef struct {
	PetscInt nCells[3];
	PetscInt lowerCorner[3];
	PetscInt upperCorner[3];
	PetscInt totalCells;
} MeshRegion;

/*E
  St_Patch - Structure for the definition of a patch.

  A patch is a rectangular region in computational domain, which is
  used, among other things, for boundary conditions and properties
  definition.

  The patch is identified by a unique patch ID, and characterized by
  the global mesh region occupied. Moreover, the patch is distributed
  among the processors by function CreatePatch, and this structure
  provides also variables for the mesh region local to each processor
  (with and without ghost nodes).

  The structure has also a pointer to the next patch.

  Attributes:
+ id - Patch identifier.
. DA - Distributed array for the patch management.
. x, y, z - Lower corner.
. nx, ny, nz - Number of elements along each coordinate.
. comm - Communicator for the group of processes owning the patch.
- Next - Pointer to the next patch in the list.

  Level: advanced
E*/
typedef struct St_Patch {
	PetscInt id;
	DA da;
	PetscInt x, y, z, nx, ny, nz;
	MPI_Comm comm;
	struct St_Patch *Next;
} St_Patch;

typedef St_Patch *Ptr_Patch;

/*E
  St_Prop - Structure for the definition of a patch-wise property.

  A patch-wise property is a constant or a function defined over a
  patch. A variable-property can depend on the geometry only, or
  on other variables or transient terms.

  When the property is variable, the structure provides a
  pointer to an array to store the values, and to a function to
  evaluate them.

  Attributes:
+ id - Property identifier.
. Patch - Patch where the property is defined.
. ktype - Property type (see ktypeProp).
. kip - Interpolation type for the value in the faces of the cells
        (see ktypeIp).
. FluxLimiter - Flux limiter.
. linrlx - Linear relaxation of the property. Not automatic; it has to be
           specified when the evaluation function is set.
. Cons - Value of the property when constant.
. Funct - Function for the evaluation of the property (when not constant).
. array - Array where the values of the property are stored.
. instantMeans - Vector for the means of the property instant values.
. instantSums - Vector for the sum of all property instant values.
. varianceMeans - Vector for the means of the property variance values.
. values - Vector for the property instant values.
. numReferences - Number of references (variables or boundary conditions
                  using this property).
- Next - Pointer to the next structure in the list.

  Level: advanced
E*/
typedef struct St_Prop {
	PetscInt id;
	Ptr_Patch Patch;
	ktypeProp ktype;
	StorageType storage;
	ktypeIp kip;
	PetscErrorCode (*FluxLimiter)(PetscInt icell, PetscInt ineighb,
	    PetscReal upwdelta, PetscReal dowdelta, PetscReal *value);
	PetscReal linrlx;
	PetscReal value;
	PetscErrorCode (*Funct)(struct St_Prop *Prop);
  PetscReal *array;
	Vec instantMeans, instantSums, varianceSums, values;
	PetscInt numReferences;
	struct St_Prop *Next;
} St_Prop;

typedef St_Prop *Ptr_Prop;

/*E
  St_BcondApplication - Structure with parameters of a boundary condition
  application.

  A boundary condition application is defined by the equation and
  variable indexes affected by the boundary condition and the properties
  needed for the evaluation. Thus, a boundary condition (St_Bcond) is
  formed by one or more applications (St_BcondApplication).

  Attributes:
+ ivar - Index of the variable to which the boundary condition applies.
. ieq - Index of the equation to which the boundary condition applies.
. CoeffA - The coefficient C in the linear source C(V - Phi)
. CoeffB - The CV product in the linear source C(V - Phi)
- next - Pointer to the next structure in the list.

  Level: advanced

.seealso St_Bcond
E*/
typedef struct St_BcondApplication {
	PetscInt ivar, ieq;
	Ptr_Prop CoeffA, CoeffB;
	struct St_BcondApplication *next;
} St_BcondApplication;

/*E
  St_Bcond - Structure with parameters of a boundary condition.

  Attributes:
+ id - Boundary condition identifier.
. ineighb - Neighbor index.
. Patch - Patch where the boundary condition is defined.
. bCondType - Boundary condition type (see BcondType).
. applicationList - List of applications of the boundary condition.
. Next - Pointer to the next structure in the list.

  Level: advanced

.seealso St_BcondApplication
E*/
typedef struct St_Bcond {
	PetscInt id, ineighb;
	Ptr_Patch Patch;
	BcondType bCondType;
	St_BcondApplication *applicationList;
	struct St_Bcond *Next;
} St_Bcond;

typedef St_Bcond *Ptr_Bcond;

/*E
  St_Geom - Structure with geometric properties of the mesh.

  Attributes:
+ Surface - Surface area of the faces of the local cells.
. Edge - Length of the finite volume in each direction.
. Volume - Cell volume.
. wallDistance - Distance to the nearest wall;
- normalVec - Unit vector, normal to the direction of the
  vector from each cell to the nearest wall cell and following
  the flow motion (i.e: trying to maximize X component and minimize
  Y component).

  Level: advanced
E*/
typedef struct {
	PetscReal *Surface[3], *Edge[3], *Volume;
	PetscReal *wallDistance, *normalVec[3];
} St_Geom;

/*E
  St_Porosity - Structure with porosity parameters.

  Attributes:
+ id - Porosity identifier.
. ineighb - Face to apply the porosity.
. Patch - Patch where the porosity is defined.
. Cons - Porosity value.
- Next - Pointer to the next structure in the list.

  Level: advanced
E*/
typedef struct St_Porosity {
	PetscInt id, ineighb;
	Ptr_Patch Patch;
	PetscReal Cons;
	struct St_Porosity *Next;
} St_Porosity;

typedef St_Porosity *Ptr_Porosity;

/*E
  St_Var - Structure with parameters of a dependent variable.

  Attributes:
+ name - Name of the variable.
. kip - Interpolation type to the value in the faces of the cells
        (see ktypeIp).
. FluxLimiter - Flux limiter.
. valmin, valmax - Minimum and maximum values of the variable.
. linrlx, fdtrlx - Relaxation values of the variable.
. resref, corref - Reference values of the variable (residue and
                   correction vector).
. Valini - Property for the initial values of the variable.
. Gamma - Property for the diffusion coefficient of the variable.
. locPhi - Local unknown vector.
. Array - Pointer to the array of locPhi vector object.
. resnorm, corrnorm - Norm values of the variable (residue and
                      correction vector).
- next - Pointer to the next structure in the list.

  Level: advanced
E*/
typedef struct St_Var {
	char name[128];
	PetscInt index;
	ktypeIp kip;
	PetscErrorCode (*FluxLimiter)(PetscInt icell, PetscInt ineighb,
	    PetscReal upwdelta, PetscReal dowdelta, PetscReal *value);
	PetscReal valmin, valmax, linrlx, fdtrlx, resref, corref,
	          resnorm, corrnorm;
	Ptr_Prop Valini, Gamma;
	Vec locPhi;
	PetscReal *Array;
	struct St_Var *next;
} St_Var;

typedef St_Var *Ptr_Var;

/*E
  St_Sys - Structure with PETSc objects for the distributed linear
  system. Different instances of the same structure are used for the
  coupled and segregated systems.

  Attributes:
+ Phi - Unknown vector.
. B - Right side vector.
. A - Coefficient matrix.
. da - Distributed array.
. solver - Linear solver.
. ltog - Local to global indexes of the cells.
. ndof - Degrees of freedom (Input.ncoupvar for coupled system, 1 for
         segregated system).
. ivar - Index of the first variable of the system. The last variable
         is ivar + ndof - 1
. resnorm - Norm of the residue.
. resref - Reference value for the residue.
. corrnorm - Norm of the correction vector.
- corref - Reference value for the correction vector.

  Level: advanced
E*/
typedef struct {
	Vec Phi, B;
	Mat A;
	DA da;
	KSP solver;
	PetscInt *ltog;
	PetscInt ndof;
	PetscInt ivar;
	PetscReal resnorm, resref, corrnorm, corref;
} St_Sys;

/*E
  St_Local - Structure with local parameters of the problem.

  Attributes:
+ rank - Processor's rank.
. meshRegion - Local subdomain.
. ghostMeshRegion - Local subdomain (including ghost nodes).
. icell_neighb - Local indexes of each neighbor for each cell.
. porosity - Porosity of each face for each cell.
- FirstPorosity, LastPorosity - Pointers to the first and last
                                porosities in the list.

  Level: advanced
E*/
typedef struct {
	PetscMPIInt rank;
	MeshRegion meshRegion, ghostMeshRegion;
	PetscInt **icell_neighb;
	PetscReal **porosity;
	St_Porosity *FirstPorosity, *LastPorosity;
} St_Local;

/*E
  St_Interp - Structure with parameters to interpolate values from
  nodes to faces, according to the convective scheme (lineal,
  upwind). It is necessary to re-calculate them after each outer
  iteration since they depend on the flow direction, except for
  the linear interpolation.

  Attributes:
+ linear - Coefficients for the linear interpolation.
. upwind - Coefficients for the upwind interpolation.
. icell_upw - Local index of the upwind cell, referring to the
              cell (notation from the paper about Flux Limiter
              (Waterson).
. icell_uupw - Local index of the up-upwind cell, referring to the
               cell (notation from the paper about Flux Limiter
               (Waterson).
. icell_dow - Local index of the downstream cell, referring to the
              cell (notation from the paper about Flux Limiter
              (Waterson).

  Level: advanced
E*/
typedef struct {
	PetscReal **linear, **upwind;
	PetscInt **icell_upw, **icell_uupw, **icell_dow;
} St_Interp;

/*E
  St_Global - Structure for the global parameters of the problem.

  Attributes:
+ comm - MPI communicator.
. NPatch - Global number of patches.
. NBcond - Global number of boundary conditions.
. NPorosity - Global number of porosities.
. NProp - Global number of properties.
. FirstPatch, LastPatch - Pointer to the first and last patches
                          of the list.
. FirstBcond, LastBcond - Pointer to the first and last boundary
                          conditions of the list.
. FirstProp, LastProp - Pointer to the first and last properties
                          of the list.

  Level: advanced
E*/
typedef struct {
	MPI_Comm comm;
	PetscInt NPatch, NBcond, NPorosity, NProp;
	St_Patch *FirstPatch, *LastPatch;
	St_Bcond *FirstBcond, *LastBcond;
	St_Prop *FirstProp, *LastProp;
} St_Global;

#endif
